__version__ = '0.5.0'
from mrktmix.data_create import aggregate_rows
from mrktmix.data_create import create_mdldata
from mrktmix.data_create import parse_variable
from mrktmix.data_create import reserve_dict
from mrktmix.data_create import spread_notna
from mrktmix.data_create import update_description
from mrktmix.optimization import mmm_optimize
from mrktmix.transformation import aggregate_data
from mrktmix.transformation import apply_apl
from mrktmix.transformation import apply_coef
from mrktmix.transformation import assess_error
from mrktmix.transformation import collapse_date
from mrktmix.transformation import create_base
from mrktmix.transformation import segregate_data



import datetime

import numpy as np
import pandas as pd
from scipy.ndimage.interpolation import shift
from scipy.signal import lfilter


def apply_apl_(input_list, adstock, power, lag):
    """ Apply advertisement decay (carry over effect or decay effect), diminishing return and lag on array

    :param input_list: list of float with marketing or any other activities like spend
    :type input_list: list
    :param adstock: Adstock, carry over effect or decay effect on activity
    :type adstock: float
    :param power: Diminishing return or power transformation on activity
    :type power: float
    :param lag: Lag transformation on activity
    :type lag: float
    :return: list after applying adstock, power and lag transformation
    :rtype: list
    """

    return(shift(np.nan_to_num(np.power(lfilter([1], [1, -float(adstock)], input_list, axis=0), power)), lag, cval=0, order=1))


def apply_apl_series(df_series, adstock, power, lag):
    """ Apply advertisement decay (carry over effect or decay effect), diminishing return and lag on pandas.Series

    :param df_series: Series with marketing or any other activities like spend
    :type df_series: pandas.Series
    :param adstock: Adstock, carry over effect or decay effect on activity
    :type adstock: float
    :param power: Diminishing return or power transformation on activity
    :type power: float
    :param lag: Lag transformation on activity
    :type lag: float
    :return: Returns input after applying adstock, power and lag transformation
    :rtype: pandas.Series
    """
    if isinstance(df_series, pd.Series):
        return(pd.Series(apply_apl_(df_series, adstock, power, lag), index=df_series.index, name=(df_series.name, adstock, power, lag)))


def create_base_(variable, date_input, freq, increasing=False, negative=False, periods=1, panel=None):
    """ Create dummy/base variable for modeling
    :param variable: Name of variable
    :type variable: str
    :param date_input: Dates used in base variable. If input type is str, periods and freq will be used to determine length of date.If
     input type is list, these dates will be used for base. If input type is tuple, the dates in the tuple will behave like range for date
    :type date_input: Union[str, list, tuple]
    :param freq: Frequency strings can have multiples, e.g. ‘5H’. See here for a list of pandas frequency aliases
    :type freq: str or DateOffset, default ‘D’
    :param increasing: Base values will in increaseing order if True. If false, it will be constant, defaults to False
    :type increasing: bool, optional
    :param negative: Controls the sign for base value, defaults to False
    :type negative: bool, optional
    :param periods: Controls number of period dates will be genrated. Will only be applicable when date_input is str, defaults to 1
    :type periods: int, optional
    :param panel: Create multiindex with panel on level 0, defaults to None. If list is supplied, it will create for each panel
    :type panel: str or list, optional
    :return: pandas Series with date and panel index
    :rtype: pandas.Series
    """
    if isinstance(date_input, tuple):
        base_df_index = pd.date_range(start=date_input[0],
                                      end=date_input[1],
                                      freq=freq)
    elif isinstance(date_input, (str, datetime.date)):
        base_df_index = pd.date_range(start=date_input,
                                      periods=periods,
                                      freq=freq)
    else:
        base_df_index = pd.to_datetime([i for i in date_input if not pd.isnull(i)]).unique()

    if panel is not None:
        if isinstance(panel, list):
            base_df_index = pd.MultiIndex.from_product([panel, base_df_index])
        else:
            base_df_index = pd.MultiIndex.from_tuples(list(zip([panel] * len(base_df_index), base_df_index)))
    base_df_index.freq = None

    base_df = pd.Series(1, index=base_df_index, name=variable)
    if increasing:
        base_df = base_df.cumsum()
    if negative:
        base_df = base_df * -1
    return(base_df)


def segregate_variable(aggregated_data, segregated_data, match_sum=True):
    """
    Segregate aggregate variable in aggregated data based on proportion in segregated data. It assumes row index in aggregated
    and segregated data are at same level

    :param aggregated_data: Aggregated series with aggregate variable
    :type aggregated_data: pandas.Series
    :param segregated_data: Segregated data is used to compute proportion for segregation of aggregated data
    :type segregated_data: pandas.DataFrame
    :param match_sum: When proportion calculated in segregated data is 0 and corresponding value is present in aggregated data,
    then sum of aggregated data won't match segregated output data. If match_sum is True, values in segregated output data is
    adjusted to match its sum with sum of aggregated data
    :type match_sum: bool
    :return: segregated data at variable level
    :rtype: pandas.DataFrame
    """

    prop_data = segregated_data.div(segregated_data.sum(axis=1), axis=0)
    segregate_df = prop_data.mul(aggregated_data, axis=0)
    if match_sum:
        adj_total = (aggregated_data - segregate_df.sum(axis=1)).sum()
        segregate_df_final = (segregate_df + segregate_df / (segregate_df.sum().sum()) * adj_total)
    else:
        segregate_df_final = segregate_df
    return(segregate_df_final)


def segregate_panel(aggregated_data, segregated_data, match_sum=True):
    """
    Segregate aggregate data at panel level based on proportion in segregated data. It assumes row index in aggregated and segregated data
    is same

    :param aggregated_data: Aggregated data at panel level
    :type aggregated_data: pandas.DataFrane
    :param segregated_data: Segregated data is used to compute proportion for segregation of aggregated data
    :type segregated_data: pandas.DataFrame
    :param match_sum: When proportion calculated in segregated data is 0 and corresponding value is present in aggregated data,
    then sum of aggregated data won't match segregated output data. If match_sum is True, values in segregated output data is
    adjusted to match its sum with sum of aggregated data
    :type match_sum: bool
    :return: segregated data at panel level
    :rtype: pandas.DataFrame
    """

    aggregate_df = segregated_data.sum(level=-1)
    prop_data = segregated_data.divide(aggregate_df.reindex(segregated_data.index, level=-1), axis=0)
    segregate_df = aggregated_data.reindex(prop_data.index, level=-1).mul(prop_data)
    if match_sum:
        adj_series = (aggregated_data.sum() - segregate_df.sum())
        segregate_df_final = (segregate_df / segregate_df.sum()).mul(adj_series).add(segregate_df)
    else:
        segregate_df_final = segregate_df
    return(segregate_df_final)



from collections import defaultdict

import numpy as np
import pandas as pd

from mrktmix.dataprep import mdl_data as dc


def create_mdldata(
        input_files,
        panel_variables,
        description_variables,
        metric_value,
        metric_index=-1,
        description2code={},
        summary_type={},
        delimeter="_",
        code_length=3,
        case_sensitive=False,
        iteratively=False,
        verbose=False):
    """
    Covert list of input files into modeling data.

    :param input_files: dictionary with filename and data
    :type input_files: dictionary with filename as key and pandas.DataFrame as values
    :param panel_variables: name of panel variables in mapped data
    :type panel_variables: list of string
    :param description_variables: name of description variables in mapped data
    :type description_variables: list of string
    :param metric_value: name of metric value (summary variable) in mapped data
    :type metric_value: string
    :param metric_index: index representing metric variable in description variables
    :type metric_index: int
    :param description2code: mapping in dictionary format with description as key and code as value, defult is empty dictionary
    :type description2code: dictionary
    :param summary_type: summary function to be applied on metric variable. key of summary type is function and
     value of summary type is list of metric values. default value is empty
    :type summary_type: dictionary
    :param delimeter: delimeter used inbetween description mapping code
    :type delimeter: string
    :param code_length: length of code to be geneated, defualt value is 3
    :type code_length: int
    :param case_sensitive: whether mapping to code is case sensitive or not
    :type case_sensitive: bool
    :param iteratively: whether modeling data processing should be done filewise or not (append data and process it in single pass)
    :type iteratively: bool
    :return: modeling dataframe with panel variable as row index and variable_name as variable. Second item of tuple represents
    mapping from description to code. Third item of tuple represents mapping file (description of variable)
    :rtype: tuples of pandas.DataFrame, dictionary, pandas.DataFrame
    """

    description2code_agg = description2code.copy()
    file_mapping = pd.DataFrame()
    if iteratively:
        mdl_data = pd.DataFrame()
        for name_df, input_df in input_files.items():
            if verbose:
                print(name_df)
            mapped_data, des2code = dc.apply_mapping(
                input_df,
                description_variables,
                code_length=code_length,
                delimeter=delimeter, description2code=description2code_agg,
                case_sensitive=case_sensitive)
            description2code_agg.update(des2code)
            mapped_data["_File_"] = name_df
            file_mapping = file_mapping.append((
                mapped_data[["_File_", "_Variable_", *description_variables]]
                .drop_duplicates()
                .set_index(["_File_", "_Variable_"])))
            mdl_df = dc.long2wide(
                mapped_data,
                panel_variables,
                description_variables[metric_index],
                metric_value,
                summary_type,
                variable_name='_Variable_',
                case_sensitive=case_sensitive)
            mdl_data = pd.concat([mdl_data, mdl_df], axis=1)
    else:
        input_df = pd.concat(input_files)
        input_df = input_df.reset_index(level=0).rename(columns={'level_0': '_File_'})
        mapped_data, des2code = dc.apply_mapping(input_df, description_variables, code_length=code_length,
                                                 delimeter=delimeter, description2code=description2code_agg, case_sensitive=case_sensitive)
        description2code_agg.update(des2code)
        file_mapping = file_mapping.append(mapped_data[["_File_", "_Variable_", *description_variables]].drop_duplicates())
        file_mapping = file_mapping.set_index(["_File_", "_Variable_"])
        mdl_data = dc.long2wide(
            mapped_data,
            panel_variables,
            description_variables[metric_index],
            metric_value,
            summary_type,
            variable_name='_Variable_',
            case_sensitive=case_sensitive)
    return(mdl_data, description2code_agg, file_mapping)


def spread_notna(input_df, prior=True):
    """
    Distributes non missing values in data frame to prior missing or post missing values equally

    :param input_df: input dataframe where adjustment needs to made
    :type input_df: pandas.DataFrame
    :param prior: whether prior or post adjustment needs to made. If prior is Ture, adjustment of non missing is spread toward
    preceding missing values. If its false, then non missing is spread toward succeeding missing values.
    :type prior: bool
    :return: Adjusted dataframe
    :rtype: pandas.DataFrame
    """

    if prior:
        shift_factor = 1
    else:
        shift_factor = -1
    change_track = input_df.isnull()
    adj_factor = (change_track.ne(change_track.shift(shift_factor))
                  .cumsum()
                  .apply(lambda x: x.map(x.value_counts()))
                  .where(change_track)
                  .shift(shift_factor)
                  .add(1)
                  .fillna(1))
    if prior:
        return(input_df.div(adj_factor).bfill())
    else:
        return(input_df.div(adj_factor).ffill())


def parse_variable(Name_series, ids_index, delimiter="_", anti=False):
    """
    Get new id/code based on id_index after splitting on given delimeter.

    :param Name_series: pandas series values will be splitted based on delimiter to find relevent ids index
    :type Name_series: pandas.Series
    :param ids_index: index in Name_series to be extracted after splitting by delimiter
    :type ids_index: list of integer
    :param delimiter: delimiter used in Name_series
    :type delimiter: string
    :param anti: index representing metric variable in description variables
    :type anti: bool
    :return: return matching ids after spliting by delimiter on Name_series
    :rtype: pandas.Series
    """

    if max(ids_index) < 0:
        var_split = Name_series.str.rsplit(pat=delimiter, n=max(map(abs, ids_index)), expand=True)
    else:
        var_split = Name_series.str.split(pat=delimiter, expand=True)
    var_split[var_split.isna()] = ""
    if anti:
        var_split['Variable'] = var_split.drop(var_split.columns[ids_index], axis=1).apply(lambda x: delimiter.join(x), axis=1)
    else:
        var_split['Variable'] = var_split.filter(var_split.columns[ids_index], axis=1).apply(lambda x: delimiter.join(x), axis=1)
    var_split['Variable'] = var_split['Variable'].str.rstrip(delimiter)
    return (var_split['Variable'])


def aggregate_rows(Dataset, metric2smry_func, metric_column, value_column, case_senstive=False, print_suffix="\t"):
    """
    Returns the data frame file after removing all the duplicates & grouping it according to metric summary dictionary

    :param Dataset: input dataset to be aggregated
    :type Dataset: pandas.DataFrame
    :param metric2smry_func: dictionary with keys as metric columns and summary function as values
    :type metric2smry_func: Dictionary
    :param metric_column: Name of Metric column on which metric2smry_func will match keys for aggregation
    :type metric_column: String
    :param value_column: Numeric variable in Dataset on which summary will be applied
    :type value_column: String
    :param case_senstive: whether keys of metric2smry will case senstive match or not
    :type case_senstive: bool
    :param print_suffix: space between first column and message printed
    :type print_suffix: string
    :return: aggregated data after applying metric2smry_func on metric_column and summarise value_column
    :rtype: pandas.DataFrame
    """

    # Find if aggregation is required in data
    duplicate_ids = Dataset.duplicated(subset=[*Dataset.columns.difference([value_column])], keep=False)
    # Aggregation of Data with duplicate ids
    df_duplicates = Dataset[duplicate_ids].copy()
    if df_duplicates.shape[0] != 0:
        if not case_senstive:
            metric2smry_func.update({k.title(): v for k, v in metric2smry_func.items()})
            [metric2smry_func.update({(i, metric2smry_func[i.title()])}) for i in df_duplicates[metric_column].unique()
             if i.title() in list(metric2smry_func)]  # same code without considering case
        print(print_suffix + "Shape of duplicate record is", df_duplicates.shape)
        smry_fun_navl = ~np.isin(df_duplicates[metric_column].unique(), list(metric2smry_func.keys()))
        if sum(smry_fun_navl):
            print("List of metric summary function not available in input dictionary. Please update it.",
                  "\n\t", df_duplicates[metric_column].unique()[smry_fun_navl],)
            raise Exception('Please update mapping dictionary.')
        else:
            df_duplicates["smry_func"] = df_duplicates[metric_column].replace(metric2smry_func)
            df_duplicates = (pd.concat([df.groupby([*Dataset.columns.difference([value_column]),
                                                    "smry_func"]).agg(smry_f) for smry_f,
                                        df in df_duplicates[[*Dataset.columns.difference([value_column]),
                                                             value_column,
                                                             "smry_func"]].groupby("smry_func")],
                                       sort=False) .reset_index() .drop(["smry_func"],
                                                                        axis=1))
    return pd.concat([Dataset[~duplicate_ids], df_duplicates], sort=False)


def reserve_dict(inputdict, concatenate="|"):
    '''
    Inserve dictionary. If there are duplicate values then it concatenate string to combine values

    :param inputdict: dictionary to be reversed
    :type inputdict: dictionary
    :param concatenate: string to be used if there are duplicate values in dictionary to concatenate
    values of reversed dictionary. Default value is '|'. If concatenate is None, values will be in form of list
    :type concatenate: string
    :return: aggregated data after applying metric2smry_func on metric_column and summarise value_column
    :rtype: dictionary
    '''
    reversed_dict = defaultdict(list)
    for key, value in inputdict.items():
        reversed_dict[value].append(key)
    if concatenate is None:
        return(dict(reversed_dict))
    else:
        return({key: concatenate.join(value) for key, value in reversed_dict.items()})


def update_description(description_input, new_code, new_description, delimeter="_", index=[], old_code=[]):
    '''
    Update description on index by replacing old code by new code and new description.

    :param description_df: description file with variable name in index and description on columns
    :type description_df: pd.DataFrame
    :param new_code: new code will be used inplace of old code in variable name
    :type new_code: string
    :param new_description: new description will be used inplace of corresponding column in description_df
    :type new_description: string
    :param delimeter: delimeters used in between code in variable name
    :type delimeter: string
    :param index: index of variable where update will be applied after spliting on delimeter. Default value is empty list. When
    empty, update will apply on all index
    :type index: list of integers
    :param old_code: list of old codes present in variable which needs to be replaced
    :type old_code: list of string
    :return: tuple of dictionary representing old variable to new variable name and updated description data
    :rtype: tuple of dictionary and pd.DataFrame
    '''
    mapping = description_input.index.to_series().str.split("_", expand=True)
    col_rename = dict(zip(description_input.columns[0:mapping.shape[1]], range(0, mapping.shape[1])))
    description_df = description_input.rename(columns=col_rename)
    if len(old_code) and len(index):
        description_df[mapping.iloc[:, index].isin(old_code)] = new_description
        mapping.iloc[:, index] = mapping.iloc[:, index].replace(to_replace=old_code, value=new_code)
    elif not len(old_code):
        description_df.iloc[:, index] = new_description
        mapping.iloc[:, index] = new_code
    elif not len(index):
        description_df[mapping.isin(old_code)] = new_description
        mapping = mapping.replace(to_replace=old_code, value=new_code)
    mapping = mapping.apply(lambda x: "_".join(x), axis=1).to_dict()
    description_df = description_df.rename(columns=reserve_dict(col_rename), index=mapping)
    description_df = description_df[~description_df.index.duplicated()]
    return (mapping, description_df)



from pytest import approx

from mrktmix.dataprep import optimization as optim


def test_optmization():
    coef = [47, 75, 13, 63, 96, 25, 17]
    intial_spend = [806, 332, 173, 661, 286, 253, 978]
    power = [0.9, 0.32, 0.97, 0.53, 0.02, 0.86, 0.67]
    lb = [644.8, 265.6, 138.4, 528.8, 228.8, 202.4, 782.4]
    ub = [967.2, 398.4, 207.6, 793.2, 343.2, 303.6, 1173.6]
    optim_result = optim.optimization(coef,
                                      intial_spend,
                                      power,
                                      lb,
                                      ub,
                                      3489,
                                      contraint_type='budget')
    optimized_spend = [967.19999999, 265.6, 207.6, 733.8, 228.8, 303.6, 782.4]
    assert optim_result.x == approx(optimized_spend)
    assert optim_result.success

    optim_result = optim.optimization(coef,
                                      intial_spend,
                                      power,
                                      lb,
                                      ub,
                                      28511.75,
                                      contraint_type='revenue')
    optimized_spend = [869.72435105, 265.6, 138.4, 528.8, 228.8, 202.4, 782.4]
    assert optim_result.x == approx(optimized_spend)
    assert optim_result.success



import numpy as np
import pandas as pd

from mrktmix.data_create import parse_variable
from mrktmix.dataprep import transform as dp


def create_base(variable, date_input, freq, increasing=False, negative=False, periods=1, panel=None):
    """ Create dummy/base variable for modeling

    :param variable: Name of variables
    :type variable: list
    :param date_input: List of string, list or tuple of Dates used in base variable. If input type is str, periods and freq will be used
     to determine length of date. If input type is list, these dates will be used for base. If input type is tuple, the dates in the tuple
     will behave like range for date
    :type date_input: list
    :param freq: list of frequency strings or frequency strings. It can have multiples, e.g. ‘5H’. See here for a list of pandas frequency
     aliases
    :type freq: Union[list, str]
    :param increasing: Base values will in increaseing order if True. If false, it will be constant, defaults to False
    :type increasing: Union[list, bool], optional
    :param negative: Controls the sign for base value, defaults to False
    :type negative: Union[list,bool], optional
    :param periods: Controls number of period dates will be genrated. Will only be applicable when date_input is str, defaults to 1
    :type periods: Union[list, int], optional
    :param panel: Create multiindex with panel on level 0, defaults to None
    :type panel: Union[list,str], optional
    :return: pandas DataFrame with date and panel index
    :rtype: pandas.DataFrame
    """
    if isinstance(variable, (np.ndarray, list, pd.Series)):
        df = pd.DataFrame({"variable": variable, "date_input": date_input, "freq": freq,
                           "increasing": increasing, "negative": negative, "periods": periods, "panel": panel})
        base_df = (df.apply(lambda row: dp.create_base_(row[0],
                                                        row[1],
                                                        row[2],
                                                        increasing=row[3],
                                                        negative=row[4],
                                                        periods=row[5],
                                                        panel=row[6]),
                            axis=1) .set_index(df.iloc[:,
                                                       0].values) .sum(level=0,
                                                                       min_count=1) .transpose())
        base_df.index.freq = None
    else:
        base_df = dp.create_base_(variable, date_input, freq, increasing=increasing, negative=negative, periods=periods, panel=panel)
    return(base_df)


def aggregate_data(mdl_data, panel_agg={}, variable_agg={}, metric_index_var=[-1], metric_mean_code=[], delimeter="_"):
    """
    Aggregate modeling data based on panel level or variable level. By defult, sum is used for aggregation.If mean function
    needs to be applied in modeling data at given code/level in variable,  metric index variable and metric mean code must
    to be supplied. The function can only one value being mapped to aggregate panel or aggregate variable. metric mean code
    is matched after variable aggregation

    :param mdl_data: modeling data with panel level and columns. Panel must be represented as multiindex rows.
    :type mdl_data: pandas.DataFrame
    :param panel_agg: Information on panel level aggregation. Keys of dicationary represents new panel name and values represents
    name of panel to be aggregated. Default value is empty dictionary.
    :type panel_agg: dictionary with list values
    :param variable_agg: Information on variable level aggregation. Keys of dicationary represents new variable and values represents
    name of variables to be aggregated. Default value is empty dictionary.
    :type variable_agg: dictionary with list values
    :param metric_index_var: index for metric type in variable of modeling data after spliting variable by delimeter supplied.
    Default value is [-1]
    :type metric_index_var: list of integer
    :param metric_mean_code: list of metric where mean will be applied. mean code is searched before applying variable aggregation
    :type metric_mean_code: list of string
    :param delimeter: delimeter used in variable in modeling data
    :type delimeter: string
    :return: modeling dataframe after aggregation at panel and variable level
    :rtype: pandas.DataFrame
    """
    # check if one variable is mapped to multiple aggregate panel
    mapped_panels = [j for i in [*panel_agg.values()] for j in i]
    multiple_mapped_panels = [i for i in mapped_panels if mapped_panels.count(i) > 1]
    # check if one variable is mapped to multiple aggregate panel
    mapped_vars = [j for i in [*variable_agg.values()] for j in i]
    multiple_mapped_vars = [i for i in mapped_vars if mapped_vars.count(i) > 1]
    if len(multiple_mapped_panels) and len(multiple_mapped_vars):
        raise Exception('Panel and variable mapping contains duplicate values')
    elif (len(multiple_mapped_panels)):
        raise Exception('Panel mapping contains duplicate values')
    elif (len(multiple_mapped_vars)):
        raise Exception('Variable mapping contains duplicate values')

    # Reverse key and value of variable agg dictionary
    variable_agg_reverse = {}
    _ = [variable_agg_reverse.update({val_item: key}) for key, val in variable_agg.items() for val_item in val]
    # Reverse key and value of variable panel dictionary
    panel_agg_reverse = {}
    _ = [panel_agg_reverse.update({val_item: key}) for key, val in panel_agg.items() for val_item in val]
    # find variables for mean summary
    mdl_data_renamed = mdl_data.rename(index=panel_agg_reverse, columns=variable_agg_reverse)
    mean_vars = parse_variable(
        pd.Series(
            mdl_data_renamed.columns.get_level_values(0),
            index=mdl_data_renamed.columns.get_level_values(0)),
        metric_index_var,
        delimiter=delimeter,
        anti=False).isin(metric_mean_code)
    sum_agg = (mdl_data_renamed.loc[:, [*~np.isin(np.array(mdl_data_renamed.columns.get_level_values(0)),
                                                  np.array([*mean_vars[mean_vars.values].index]))]]
               .sum(level=list(range(mdl_data_renamed.index.nlevels)))
               .sum(axis=1, level=list(range(0, mdl_data_renamed.columns.nlevels))))

    # mean aggregation
    if len([*mean_vars[mean_vars.values].index]):
        mean_agg = (mdl_data_renamed[[*mean_vars[mean_vars.values].index]]
                    .mean(level=mdl_data_renamed.index.names)
                    .mean(axis=1, level=list(range(0, mdl_data_renamed.columns.nlevels))))
        agg_mdl_data = pd.concat([sum_agg, mean_agg], axis=1)
    else:
        agg_mdl_data = sum_agg
    return(agg_mdl_data)


def segregate_data(aggregated_data, segregated_data, panel_agg={}, variable_agg={}, panel_agg_index=0, match_sum=True):
    """
    Segregate aggregated data based on ratio in segregated data. Aggregated data can have panel level aggregation or variable
    level aggregation. If panel level aggregation is present, its mapping must be present in panel_agg. If variable level
    aggregation present, it must be present in variable_agg. Panel level aggregation is applied on panel_agg_index level
    of row index.

    :param aggregated_data: Aggregated data at panel level
    :type aggregated_data: pandas.DataFrane
    :param segregated_data: Segregated data is used to compute proportion for segregation of aggregated data
    :type segregated_data: pandas.DataFrame
    :param panel_agg: Information on panel level aggregation. Keys of dicationary represents new panel name (in aggregated data)
    and values represents name of panel (in segregated data) to be aggregated. Default value is empty dictionary.
    :type panel_agg: dictionary with list values
    :param variable_agg: Information on variable level aggregation. Keys of dicationary represents new variable (in aggregated data)
    and values represents name of variables (in segregated data) to be aggregated. Default value is empty dictionary.
    :type variable_agg: dictionary with list values
    :param match_sum: When proportion calculated in segregated data is 0 and corresponding value is present in aggregated data,
    then sum of aggregated data won't match segregated output data. If match_sum is True, values in segregated output data is
    adjusted to match its sum with sum of aggregated data
    :type match_sum: bool
    :return: segregated data at panel level
    :rtype: pandas.DataFrame
    """

    # Subset relevent variable_agg_subset
    variable_agg_subset = {}
    _ = [variable_agg_subset.update({i: j}) for i, j in variable_agg.items() if i in aggregated_data.columns]
    # Subset relevent panel_agg_subset
    panel_agg_subset = {}
    _ = [panel_agg_subset.update({i: j}) for i, j in panel_agg.items(
    ) if i in aggregated_data.index.get_level_values(panel_agg_index).unique()]

    # aggregate segregated data at panel level
    panel_level_seg_agg = (aggregate_data(segregated_data, panel_agg=panel_agg_subset, variable_agg={})
                           .reindex(aggregated_data.index))
    # Variable level segregation on panel level aggregated data
    var_level_segg = pd.DataFrame()
    for agg_var, seg_vars in variable_agg_subset.items():
        var_level_segg_temp = dp.segregate_variable(aggregated_data[agg_var], panel_level_seg_agg[seg_vars], match_sum=match_sum)
        var_level_segg = pd.concat([var_level_segg, var_level_segg_temp], axis=1)
    var_level_segg_all = pd.concat([var_level_segg, aggregated_data.loc[:, aggregated_data.columns.isin(
        panel_level_seg_agg.columns) & ~aggregated_data.columns.isin(variable_agg_subset.keys())]], axis=1)

    # Panel level segregation on variable level segregated data
    panel_level_segg = pd.DataFrame()
    for agg_panel, seg_panel in panel_agg_subset.items():
        panel_level_segg_temp = segregated_data.loc[segregated_data.index.get_level_values(
            panel_agg_index).isin(seg_panel), var_level_segg_all.columns]
        panel_level_agg = (var_level_segg_all[var_level_segg_all.index.get_level_values(panel_agg_index).isin([agg_panel])]
                           .droplevel(level=panel_agg_index))
        panel_level_segg = pd.concat([panel_level_segg, dp.segregate_panel(
            panel_level_agg, panel_level_segg_temp, match_sum=match_sum)], axis=0)
    panel_var_seg_data = pd.concat(
        [panel_level_segg,
         var_level_segg_all[~var_level_segg_all.index.get_level_values(panel_agg_index).isin([*panel_agg.keys()])]], axis=0)
    return(panel_var_seg_data)


def apply_apl(dframe, dict_apl):
    """ Apply advertisement decay (carry over effect or decay effect), diminishing return and lag on pandas.DataFrame

    :param dframe: DataFrame with marketing or any other activities like spend
    :type dframe: pandas.DataFrame
    :param dict_apl: Dictionary with list of Adstock, carry over effect or decay effect on activity. Key of the dictionary can be name of
        panel or Bool. If key is true, then transformation is applied at panel level. If key is false, then transformation is applied on
        entire dataframe. If trasformation needs to be applied at panel level (name of panel should at level -2 in multiindex row
    :type dict_apl: dictionary with list values. Keys can be string or bool
    :return: Returns input after applying adstock, power and lag transformation
    :rtype: pandas.DataFrame
    """
    if (len(dict_apl) == 1) and (False in dict_apl.keys()):
        all_vars = dict_apl[False]
        df_transformed = pd.concat([dp.apply_apl_series(dframe[var[0]], var[1], var[2], var[3]) for var in all_vars], axis=1)
    elif (len(dict_apl) == 1) and (True in dict_apl.keys()):
        all_vars = dict_apl[True]
        df_grouped = dframe.groupby(dframe.droplevel(-1).index)
        df_transformed = pd.concat([df_grouped[var[0]].transform(
            lambda x: dp.apply_apl_series(x, var[1], var[2], var[3])) for var in all_vars], axis=1)
        df_transformed.columns = pd.MultiIndex.from_tuples(all_vars)
    else:
        df_transformed = pd.DataFrame()
        for panel, all_vars in dict_apl.items():
            df_transformed = pd.concat([df_transformed, pd.concat([dp.apply_apl_series(
                dframe.loc[[panel], var[0]], var[1], var[2], var[3]) for var in all_vars], axis=1)])
    df_transformed.columns.names = ["Variable", "Adstock", "Power", "Lag"]
    return(df_transformed)


def apply_coef_(raw_data, coef, dep_series):
    """ Apply coefficient and transformation on raw data

    :param raw_data: modeling dataframe with date index at level -1. If panel is present, it should be at level -2.
    :type raw_data: pandas.DataFrame
    :param coef: Coefficient and parameter to be applied on modeling dataframe. Coefficient should have tuple of variable, adstock,power
        and lag at index level -1. If panel is present in modeling data, then coefficient must have panel information in index at level -2.
    :type coef: pandas.Series
    :param dep_series: Dependent Series will be used to calculate residual. It must be at same level modeling dataframe
    :type dep_series: pandas.Series or None
    :return: Decomposition of dependent series
    :rtype: pandas.DataFrame
    """
    if coef.index.nlevels != raw_data.index.nlevels:
        raise Exception('Mismatch of index in input data')
    if coef.index.nlevels - 1:
        panel_var = (pd.Series(coef.index.get_level_values(-1))
                     .groupby(coef.droplevel(-1).index)
                     .apply(list)
                     .to_dict())
    else:
        panel_var = {False: [*coef.index]}

    after_apl = apply_apl(raw_data, panel_var)
    if coef.index.nlevels - 1:
        coef2frame = coef.unstack()
        coef2frame.columns = pd.MultiIndex.from_tuples(coef2frame.columns)
        dep_decomposition = after_apl[coef2frame.columns].mul(coef2frame.reindex(after_apl.index, level=0))
    else:
        dep_decomposition = after_apl.mul(coef)
    if dep_series is not None:
        dep_decomposition[("Residual", 0, 1, 0)] = dep_series - dep_decomposition.sum(axis=1)
    return(dep_decomposition)


def apply_coef_node_(mdl_data, coef, nodes, node, node_split):
    """
    Apply coef on modeling data to create decomposition.

    :param mdl_data: modeling dataframe with date index at level -1. If panel is present, it should be at level -2.
    :type mdl_data: pandas.DataFrame
    :param coef: Coefficient to be applied on modeling dataframe. Coefficient should have tuple of variable, adstock,power
        and lag at index level -1. If panel is present in modeling data, then coefficient must have panel information in index at level -2.
    :type coef: pandas.Series
    :param nodes: Relationship between indepenent variable and node to be applied on modeling dataframe. Nodes should
    have tuple of independent variable, adstock,power and lag at index level -1. If panel is present in modeling data, then nodes
    must have panel information in index at level -2. Dependent variables should be as value of series
    :type nodes: pandas.Series
    :param node: Selected nodes for which decomposition will be created
    :type dep_series: string
    :param node_split: Total number of relationshipship for the given node. It will equally seperate decomposition in-between the nodes
    :type node_split: int
    :return: Decomposition of dependent series for given node
    :rtype: pandas.DataFrame
    """

    # apply apl and create model decomposition
    df_apl = apply_coef_(
        mdl_data,
        coef[nodes == node],
        apply_apl(mdl_data, {False: [node]}).iloc[:, 0])
    # Decomposition of nodes
    if node_split:
        adj_decomposition = pd.concat([
            df_apl.loc[:, df_apl.columns.get_level_values(0).isin(["Intercept", "Residual"])].sum(axis=1).rename(node),
            (df_apl.assign(_Adjustment_Factor_=lambda x: x.sum(axis=1) * -1)
             .drop(columns=["Intercept", "Residual"], level=0))], axis=1)
    else:
        adj_decomposition = df_apl
    dep_decomposition = (pd.concat([adj_decomposition], axis=1, keys=[("", node)])
                         .droplevel(level=0, axis=1))
    return(dep_decomposition)


def apply_coef(mdl_data, coef, nodes=None, dep_series=None):
    """
    Apply coef on modeling data to create decomposition. If given node is present more than one relationships, then decomposed
    series is equally divided into the independent nodes.

    :param mdl_data: modeling dataframe with date index at level -1. If panel is present, it should be at level -2.
    :type mdl_data: pandas.DataFrame
    :param coef: Coefficient to be applied on modeling dataframe. Coefficient should have tuple of variable, adstock,power
        and lag at index level -1. If panel is present in modeling data, then coefficient must have panel information in index at level -2.
    :type coef: pandas.Series
    :param nodes: Relationship between indepenent variable and node to be applied on modeling dataframe. Nodes should
    have tuple of independent variable, adstock,power and lag at index level -1. If panel is present in modeling data, then nodes
    must have panel information in index at level -2. Dependent variables should be as value of series. Default value is None
    :type nodes: pandas.Series
    :param dep_series: Dependent Series will be used to calculate residual. It must be at same level modeling dataframe. Default
    is None. If dependent is not None, then residuals will also be calculated
    :type dep_series: pandas.Series or None
    :return: Decomposition of dependent series for given node
    :rtype: pandas.DataFrame
    """
    if nodes is None:
        # no nodes are present. simple case of decomposition
        return(apply_coef_(mdl_data, coef, dep_series))
    else:
        # Panel is present
        if nodes.index.nlevels == 2:
            network_decomposition = pd.DataFrame()
            for panel in nodes.index.get_level_values(0).unique():
                # count of nodes
                nodes_count = {node: sum(coef[panel][nodes[panel] != node].index == node) for node in nodes[panel].unique()}
                panel_decomposition = pd.concat([
                    apply_coef_node_(mdl_data.loc[[(panel)], :],
                                     coef[[panel]],
                                     nodes[[panel]],
                                     node,
                                     node_split) for node, node_split in nodes_count.items()], axis=1)
                network_decomposition = pd.concat([network_decomposition, panel_decomposition])
        # Panel is not present
        else:
            # count of nodes
            nodes_count = {node: sum(coef[nodes != node].index == node) for node in nodes.unique()}
            network_decomposition = pd.concat([apply_coef_node_(mdl_data, coef, nodes, node, node_split)
                                               for node, node_split in nodes_count.items()], axis=1)
        return(network_decomposition)


def collapse_date(dep_decompose, date_dict):
    """ Summarise data after collapsing date (index at level -1). Summarization is based on date dictionary given in input.

    :param dep_decompose: data to be summarised. date should be index with level -1.
    :type dep_decompose: pandas.DataFrame
    :param date_dict: dictionary with name and tuples of dates. Dates are inclusive.
    :type date_dict: dictionary
    :return: Summarise after collpasing date
    :rtype: pandas.DataFrame
    """
    all_decomp_smry = []
    for key, val in date_dict.items():
        selected_date = [upp_lim == low_lim for upp_lim, low_lim in zip(
            val[1] >= dep_decompose.index.get_level_values(-1), dep_decompose.index.get_level_values(-1) >= val[0])]
        if dep_decompose.index.nlevels - 1:
            decomp_smry = dep_decompose.iloc[selected_date].sum(level=range(0, dep_decompose.index.nlevels - 1))
        else:
            decomp_smry = dep_decompose.iloc[selected_date].sum()
        if isinstance(decomp_smry, pd.Series):
            decomp_smry = decomp_smry.rename(key)
        else:
            decomp_smry = decomp_smry.stack(list(range(0, dep_decompose.columns.nlevels))).rename(key)
        if len(all_decomp_smry):
            all_decomp_smry = pd.concat([all_decomp_smry, decomp_smry], axis=1)
        else:
            all_decomp_smry = decomp_smry.to_frame()
    return(all_decomp_smry)


def assess_error(dep_decompose):
    """ Create actual vs predicted with error terms from given resonse decomposition

    :param dep_decompose: data with decomposition of response variable and Residual term
    :type dep_decompose: pandas.DataFrame
    :return: dataframe with actual, predicted, error and error % to measure accury of model
    :rtype: pandas.DataFrame
    """
    dep = dep_decompose.sum(axis=1).rename("Dependent")
    pred = dep_decompose.drop("Residual", axis=1, level=-4).sum(axis=1).rename("Predicted")
    residual = (dep - pred).rename("Error")
    residual_perc = (residual / dep).rename("Error %")
    return(pd.concat([dep, pred, residual, residual_perc], axis=1))



import random

import numpy as np
import pandas as pd

from mrktmix.dataprep import mdl_data as raw_dp


def test_generate_code():
    # test default value
    random.seed(999)
    assert raw_dp.generate_code(3) == '2C5'
    # test with given code length
    random.seed(999)
    assert raw_dp.generate_code(5) == '2C5UR'
    # test with given chars
    random.seed(999)
    assert raw_dp.generate_code(5, chars="ABC") == 'CACBB'


def test_apply_mapping():
    # Case sensitive = True feature
    long_format_data = {'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
                        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
                        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'}}
    long_format_data = pd.DataFrame.from_dict(long_format_data)
    output_data = {
        'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'},
        '_Variable_': {0: 'EVE_RRG', 1: 'ETE_SPE', 2: 'TVT_PGP', 3: 'TVT_PGP'}}
    output_data = pd.DataFrame.from_dict(output_data)
    output_dict = {'Spend': 'SPE', 'TV': 'TVT', 'GRP': 'PGP', 'GRp': 'RRG', 'TV_Free': 'EVE', 'TV_free': 'ETE'}
    random.seed(999)
    mapped_data, description_mapping = raw_dp.apply_mapping(
        long_format_data, ['Channel', 'Metric'], delimeter="_", description2code={}, case_sensitive=True)
    pd.testing.assert_frame_equal(mapped_data, output_data)
    assert output_dict == description_mapping

    # Case sensitive = False feature
    long_format_data = {'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
                        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
                        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'}}
    long_format_data = pd.DataFrame.from_dict(long_format_data)
    output_data = {
        'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'},
        '_Variable_': {0: 'EVE_PGP', 1: 'EVE_SPE', 2: 'TVT_PGP', 3: 'TVT_PGP'}}
    output_data = pd.DataFrame.from_dict(output_data)
    output_dict = {'Spend': 'SPE', 'TV': 'TVT', 'GRP': 'PGP', 'GRp': 'PGP', 'TV_Free': 'EVE', 'TV_free': 'EVE'}
    random.seed(999)
    mapped_data, description_mapping = raw_dp.apply_mapping(
        long_format_data, ['Channel', 'Metric'], delimeter="_", description2code={}, case_sensitive=False)
    pd.testing.assert_frame_equal(mapped_data, output_data)
    assert output_dict == description_mapping

    # Test different delimeter
    long_format_data = {'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
                        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
                        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'}}
    long_format_data = pd.DataFrame.from_dict(long_format_data)
    output_data = {
        'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'},
        '_Variable_': {0: 'EVE.PGP', 1: 'EVE.SPE', 2: 'TVT.PGP', 3: 'TVT.PGP'}}
    output_data = pd.DataFrame.from_dict(output_data)
    output_dict = {'Spend': 'SPE', 'TV': 'TVT', 'GRP': 'PGP', 'GRp': 'PGP', 'TV_Free': 'EVE', 'TV_free': 'EVE'}
    random.seed(999)
    mapped_data, description_mapping = raw_dp.apply_mapping(
        long_format_data, ['Channel', 'Metric'], delimeter=".", description2code={}, case_sensitive=False)
    pd.testing.assert_frame_equal(mapped_data, output_data)
    assert output_dict == description_mapping

    # Test predefined mapping feature when case sensitive is true
    long_format_data = {'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
                        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
                        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'}}
    long_format_data = pd.DataFrame.from_dict(long_format_data)
    output_data = {
        'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'},
        '_Variable_': {0: 'ETE.GRP', 1: 'RFT.SPE', 2: 'tv.TV_GRP', 3: 'tv.TV_GRP'}}
    output_data = pd.DataFrame.from_dict(output_data)
    output_dict = {'GRP': 'TV_GRP', 'TV': 'tv', 'GRp': 'GRP', 'Spend': 'SPE', 'TV_Free': 'ETE', 'TV_free': 'RFT'}
    random.seed(999)
    mapped_data, description_mapping = raw_dp.apply_mapping(
        long_format_data, [
            'Channel', 'Metric'], delimeter=".", description2code={
            'GRP': 'TV_GRP', "TV": "tv"}, case_sensitive=True)
    pd.testing.assert_frame_equal(mapped_data, output_data)
    assert output_dict == description_mapping

    # Test predefined mapping feature when case sensitive is false
    long_format_data = {'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
                        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
                        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'}}
    long_format_data = pd.DataFrame.from_dict(long_format_data)
    output_data = {
        'Panel': {0: 'panel1', 1: 'panel1', 2: 'panel1', 3: 'panel1'},
        'Channel': {0: 'TV_Free', 1: 'TV_free', 2: 'TV', 3: 'TV'},
        'Metric': {0: 'GRp', 1: 'Spend', 2: 'GRP', 3: 'GRP'},
        '_Variable_': {0: 'ETE.TV_GRP', 1: 'ETE.SPE', 2: 'tv.TV_GRP', 3: 'tv.TV_GRP'}}
    output_data = pd.DataFrame.from_dict(output_data)
    output_dict = {'GRP': 'TV_GRP', 'GRp': 'TV_GRP', 'TV': 'tv', 'Spend': 'SPE', 'TV_Free': 'ETE', 'TV_free': 'ETE'}
    random.seed(999)
    mapped_data, description_mapping = raw_dp.apply_mapping(long_format_data, ['Channel', 'Metric'], delimeter=".", description2code={
                                                            'GRP': 'TV_GRP', "TV": "tv", "TV": "tv"}, case_sensitive=False)
    pd.testing.assert_frame_equal(mapped_data, output_data)
    assert output_dict == description_mapping


def test_long2wide():
    # input data
    input_data = {'Panel': {0: 'panel1',
                            1: 'panel1',
                            2: 'panel2',
                            3: 'panel2',
                            4: 'panel2',
                            5: 'panel3',
                            6: 'panel3'},
                  'Channel': {0: 'TV_Free',
                              1: 'TV_Free',
                              2: 'TV_Free',
                              3: 'TV_Free',
                              4: 'TV_Free',
                              5: 'TV_Free',
                              6: 'TV_Free'},
                  'Metric': {0: 'GRp',
                             1: 'Spend',
                             2: 'GRP',
                             3: 'SpenD',
                             4: 'Spend',
                             5: 'gRP',
                             6: 'gRP'},
                  'Metric_Value': {0: 33, 1: 102, 2: 45, 3: 129, 4: 170, 5: 24, 6: 49},
                  'Variable': {0: 'GRP',
                               1: 'SPD',
                               2: 'GRP',
                               3: 'SPD',
                               4: 'SPD',
                               5: 'GRP',
                               6: 'GRP'}}
    input_data = pd.DataFrame.from_dict(input_data)

    # Test when summary type is case sensitive
    output = {('mean', 'GRP'): {'panel1': 33.0, 'panel2': np.nan, 'panel3': np.nan},
              ('sum', 'GRP'): {'panel1': np.nan, 'panel2': 45.0, 'panel3': 73.0},
              ('sum', 'SPD'): {'panel1': 102.0, 'panel2': 299.0, 'panel3': np.nan}}
    output = pd.DataFrame.from_dict(output)
    output.columns.names = ["_summary_type_", 'Variable']
    output.index.names = ['Panel']
    pd.testing.assert_frame_equal(raw_dp.long2wide(input_data, ["Panel"], "Metric", 'Metric_Value', {
                                  'mean': ['GRp']}, variable_name='Variable', case_sensitive=True), output)

    # Test when summary type is not case sensitive
    output = {('mean', 'GRP'): {'panel1': 33.0, 'panel2': 45.0, 'panel3': 36.5},
              ('sum', 'SPD'): {'panel1': 102.0, 'panel2': 299.0, 'panel3': np.nan}}
    output = pd.DataFrame.from_dict(output)
    output.columns.names = ["_summary_type_", 'Variable']
    output.index.names = ['Panel']
    pd.testing.assert_frame_equal(raw_dp.long2wide(input_data, ["Panel"], "Metric", 'Metric_Value', {
                                  'mean': ['GRp']}, variable_name='Variable', case_sensitive=False), output)

    # multiple summary type for given metric
    output = {('mean', 'GRP'): {'panel1': 33.0, 'panel2': 45.0, 'panel3': 36.5},
              ('max', 'GRP'): {'panel1': 33, 'panel2': 45, 'panel3': 49},
              ('sum', 'SPD'): {'panel1': 102.0, 'panel2': 299.0, 'panel3': np.nan}}
    output = pd.DataFrame.from_dict(output)
    output.columns.names = ["_summary_type_", 'Variable']
    output.index.names = ['Panel']
    pd.testing.assert_frame_equal(raw_dp.long2wide(input_data, ["Panel"], "Metric", 'Metric_Value', {
                                  'mean': ['Grp'], 'max': ['Grp']}, variable_name='Variable', case_sensitive=False), output)

    # multiple summary type for given metric
    output = {('mean', 'GRP'): {'panel1': 33.0, 'panel2': 45.0, 'panel3': 36.5},
              ('mean', 'SPD'): {'panel1': 102.0, 'panel2': 149.5, 'panel3': np.nan},
              ('max', 'GRP'): {'panel1': 33, 'panel2': 45, 'panel3': 49},
              ('sum', 'SPD'): {'panel1': 102.0, 'panel2': 299.0, 'panel3': np.nan}}
    output = pd.DataFrame.from_dict(output)
    output.columns.names = ["_summary_type_", 'Variable']
    output.index.names = ['Panel']
    pd.testing.assert_frame_equal(raw_dp.long2wide(input_data, ["Panel"], "Metric", 'Metric_Value', {'mean': ['Grp', "Spend"], 'max': [
                                  'Grp'], 'sum': ["Spend"]}, variable_name='Variable', case_sensitive=False), output)

    # multiple summary type for given metric
    output = {('mean', 'GRP'): {'panel1': 33.0, 'panel2': 45.0, 'panel3': 36.5},
              ('mean', 'SPD'): {'panel1': 102.0, 'panel2': 149.5, 'panel3': np.nan},
              ('max', 'GRP'): {'panel1': 33, 'panel2': 45, 'panel3': 49}}
    output = pd.DataFrame.from_dict(output)
    output.columns.names = ["_summary_type_", 'Variable']
    output.index.names = ['Panel']
    pd.testing.assert_frame_equal(
        raw_dp.long2wide(
            input_data, ["Panel"], "Metric", 'Metric_Value', {
                'mean': [
                    'Grp', "Spend"], 'max': ['Grp']}, variable_name='Variable', case_sensitive=False), output)



import numpy as np
import pandas as pd
from scipy.optimize import Bounds
from scipy.optimize import minimize


def optimization(_coef,
                 _init,
                 _power,
                 lower_bound,
                 upper_bound,
                 tot_constraint,
                 contraint_type="budget"):
    """ Optimize revenue/budget (Revenue= summation of coef * spend ^ power) based in initial value and constraint. Constraints
    is in form of lower bound, upper bound and either budget or revenue

    :param _coef: list of float representing coefficients
    :type _coef: list of float
    :param _init: list of float representing initial spend
    :type _init: list of float
    :param _power: list of float representing diminshing return on spend
    :type _power: list of float
    :param lower_bound: list of number representing lower bound for spend
    :type lower_bound: list of number
    :param upper_bound: list of number representing upper bound for spend
    :type upper_bound: list of number
    :param tot_constraint: total contraint either total spend or total revenue
    :type tot_constraint: numeric
    :param contraint_type: defines type for tot_constraint. Default value is 'budget'
    :type contraint_type: string, optional
    :return: Optimized result
    :rtype: scipy.optimize.optimize.OptimizeResult
    """

    _coef = pd.Series(_coef)
    _power = pd.Series(_power)
    _init = pd.Series(_init)

    if contraint_type == "revenue":
        def objective(x):
            return np.sum(x)

        def obj_der(x):
            return pd.Series([1] * len(x))

        def constrain(x, tot_constraint=tot_constraint, _coef=_coef, _power=_power):
            return tot_constraint - np.sum(_coef * (x**_power))
    else:
        def objective(x, _coef=_coef, _power=_power):
            return -np.sum(_coef * (x**_power))

        def obj_der(x, _coef=_coef, _power=_power):
            return -_coef * _power * (x**(_power - 1))

        def constrain(x, tot_constraint=tot_constraint):
            return tot_constraint - sum(x)
    return (minimize(fun=objective, x0=_init, jac=obj_der, method='SLSQP',  # trust-constr,SLSQP
                     bounds=Bounds(lower_bound, upper_bound),
                     constraints=[{'type': 'eq', 'fun': constrain}]))



import random

import numpy as np
import pandas as pd


def generate_code(code_length, chars='ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    """
    Generate random code from ascii and digits based character length

    :param code_length: length of code to be generated.
    :type code_length: int
    :param chars: characters from which random literals will be selected
    :type chars: string
    :return: random slected literal of length given in code length from chars
    :rtype: string
    """
    return (''.join(random.choices(chars, k=code_length)))


def apply_mapping(long_format_data, description_variables, code_length=3, delimeter="_", description2code={}, case_sensitive=False):
    """
    Create/apply mapping on description variable on data to create variable name for modeling data

    :param long_format_data: dataframe to be used to create modeling data
    :type long_format_data: pandas.DataFrame
    :param description_variables: columns in dataframe for which mapping codes needs to generated
    :type description_variables: list of string
    :param code_length: length of code to be geneated, defualt value is 3
    :type code_length: int
    :param delimeter: delimeter used inbetween description mapping code
    :type delimeter: string
    :param description2code: mapping in dictionary format with description as key and code as value, defult is empty dictionary
    :type description2code: dictionary
    :param case_sensitive: whether mapping to code is case sensitive or not
    :type case_sensitive: bool
    :return: dataframe after applying mapping code with column name "_Variable_". Also return mapping of description and code
    :rtype: tuple of pandas.DataFrame and dictionary
    """

    df4mdl = long_format_data.copy()
    # Description present in supplied mapping dictionary
    new_description = np.unique(df4mdl[description_variables].values)
    # subset relevant mapping
    des2code = {}
    # same code without considering case
    if not case_sensitive:
        description2code_title = {k.title(): v for k, v in description2code.items()}
        [des2code.update({(i, description2code_title[i.title()])}) for i in new_description if i.title() in list(description2code_title)]
    # update dictionary if case matches
    des2code.update({k: description2code[k] for k in new_description[np.isin(new_description, list(description2code.keys()))]})

    # Generate new mapping if description is not present in supplied mapping dictionary
    new_description = new_description[~np.isin(new_description, list(des2code.keys()))]
    des2code_keep = {}
    if len(new_description):
        des2code_new = {}
        avoid_infinite_loop = 1
        while True:
            # generate code
            series4code = pd.Series(new_description, index=new_description).replace(r'\W|_', '', regex=True)
            if avoid_infinite_loop == 1:
                des2code_generate = (series4code * ((code_length / series4code.str.len()).apply(np.ceil)
                                                    ).astype(int)).str[:code_length].str.upper()
            elif avoid_infinite_loop == 2:
                des2code_generate = series4code.apply(lambda x: generate_code(code_length, x.upper()))
            elif avoid_infinite_loop == 3:
                des2code_generate = series4code.apply(lambda x: generate_code(code_length, "ABCDEFGHIJKLMNOPQRSTUVWXYZ"))
            else:
                des2code_generate = series4code.apply(lambda x: generate_code(code_length, "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"))
            # check if code is duplicate or exists in original description to code mapping
            des_code_dup = des2code_generate.isin(des2code.values()) | des2code_generate.duplicated(keep=False)
            # Finalize mapping if duplicate is not present
            des2code_new.update(des2code_generate[~des_code_dup].to_dict())
            # update new description if duplicate codes are present
            new_description = des_code_dup[des_code_dup].index
            avoid_infinite_loop = avoid_infinite_loop + 1
            # break loop if no new description is present
            if not(des_code_dup.sum()):
                break
        if not case_sensitive:
            # Code should be common if description is matches after ignoring case
            des2code_new = dict([(x, v) if x.lower() not in des2code_keep and not des2code_keep.update(
                {x.lower(): v}) else (x, des2code_keep[x.lower()]) for x, v in des2code_new.items()])
        des2code.update(des2code_new)

    # Create Variable in raw file
    df4mdl["_Variable_"] = (df4mdl[description_variables]
                            .replace(des2code)
                            .apply(lambda x: delimeter.join(x), axis=1))
    return(df4mdl, des2code)


def long2wide(mapped_data, panel_variables, metric_var, value_variable, summary_type={}, variable_name='_Variable_', case_sensitive=False):
    """
    Covert long format data into modeling data.

    :param mapped_data: dataframe to be converted to modeling data
    :type mapped_data: pandas.DataFrame
    :param panel_variables: name of panel variables in mapped data
    :type panel_variables: list of string
    :param metric_var: name of metric variables in mapped data
    :type metric_var: string
    :param metric_value: name of metric value (summary variable) in mapped data
    :type metric_value: string
    :param summary_type: summary function to be applied on metric variable. key of summary type is function and
     value of summary type is list of metric values. default value is empty
    :type summary_type: dictionary
    :param variable_name: variable name in mapped data to be mapped to modeling data
    :type variable_name: string
    :param case_sensitive: whether mapping to code is case sensitive or not
    :type case_sensitive: bool
    :return: modeling dataframe with panel variable as row index and variable_name as variable
    :rtype: pandas.DataFrame
    """

    df4mdl = mapped_data.copy()
    if not case_sensitive:
        # convert all metric to title case
        df4mdl[metric_var] = df4mdl[metric_var].str.title()
        # convert all values of dictionary to title case
        summary_type = {k: [i.title() for i in v] for k, v in summary_type.items()}
    # modeling data
    mdl_data = pd.DataFrame()
    # modeling data where summary type is present
    for k, v in summary_type.items():
        mdl_df = pd.pivot_table(df4mdl[df4mdl[metric_var].isin(v)].assign(_summary_type_=k),
                                values=value_variable,
                                index=panel_variables,
                                columns=["_summary_type_", variable_name],
                                aggfunc=k)
        mdl_data = pd.concat([mdl_data, mdl_df], axis=1)
    # modeling data where summary type is not present
    mdl_df = pd.pivot_table(df4mdl[~df4mdl[metric_var].isin([j for i in summary_type.values() for j in i])].assign(_summary_type_='sum'),
                            values=value_variable,
                            index=panel_variables,
                            columns=["_summary_type_", variable_name],
                            aggfunc='sum')
    mdl_data = pd.concat([mdl_data, mdl_df], axis=1)
    mdl_data.index.names = mdl_df.index.names
    return(mdl_data)
